/*
 * msg.c - routines for error messages
 */

/* Copyright  1986, 1988, 1989 1991 the Free Software Foundation, Inc.
 * 		This file is part of GAWK, the GNU implementation of the
 * AWK Progamming Language, modified for the Macintosh (also called hAWK).
 * 		GAWK is free software; you can redistribute or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or any later version.
 * 		GAWK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 		You should have received a copy of the GNU General Public License
 * along with GAWK; see the file "COPYING hAWK". If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 * Modified for THINK C 4 on the Macintosh by Ken Earle (Dynabyte) Aug 1991.
 */

#include "AWK.H"

/* Mac note, error message routines are not that great at the
moment (ie barely adequate), this will improve when time permits. */

static void err(char *s, char *msg, va_list *argp);
void msg(char *va_alist, ...);
void warning(char *va_alist, ...);
void fatal(char *va_alist, ...);

/* these two are reinited in main */
short sourceline = 0;
char *source = NULL;

/* VARARGS2 */
static void err(char *s, char *msg, va_list *argp)
{
	short line;
	char *file;

	(void) fprintf(stderr, "%s: %s ", myname, s);
	vfprintf(stderr, msg, *argp);
	(void) fprintf(stderr, "\n");
	line = (short) FNR_node->var_value->numbr;
	if (line) {
		(void) fprintf(stderr, " input line number %d", line);
		file = FILENAME_node->var_value->stptr;
		if (file && !STREQ(file, "-"))
			(void) fprintf(stderr, ", file `%s'", file);
		(void) fprintf(stderr, "\n");
	}

	if (sourceline) {
		(void) fprintf(stderr, " source line number %d", sourceline);
		if (source)
			(void) fprintf(stderr, ", file `%s'", source);
		(void) fprintf(stderr, "\n");
	}

}

/*VARARGS0*/
void msg(char *va_alist, ...)
{
	va_list args;
	//char *mesg;

	va_start(args, va_alist);
	//mesg = va_arg(args, char *);
	//err("", mesg, &args);
	err("", va_alist, &args);
	va_end(args);
}

/*VARARGS0*/
void warning(char *va_alist, ...)
{
	va_list args;
	//char *mesg;

	va_start(args, va_alist);
	//mesg = va_arg(args, char *);
	//err("warning:", mesg, &args);
	err("warning:", va_alist, &args);
	va_end(args);
}

/*VARARGS0*/
void fatal(char *va_alist, ...)
{
	va_list args;
	//char *mesg;

	va_start(args, va_alist);
	//mesg = va_arg(args, char *);
	//err("fatal error:", mesg, &args);
	err("fatal error:", va_alist, &args);
	va_end(args);
#ifdef DEBUG
	abort();
#endif
	exit(1);
}
